#Include once "fbgfx.bi"
#include Once "Inc\sGUI\sGUI.bas"
#include once "Inc\sGUI\ScrollBar.bas"
#include once "Inc\sGUI\SimpleGadget.bas"

Using FB
using sGUI
#Include Once "3dsloader.bas"
#Include once "platform.bi" '' platform specific definitions
#include once "core.bi" '' core functions and defines
#include once "math.bi" '' math functions and constants
#include once "vec4.bi" '' vec4 type and operations
#include once "mat4.bi" '' mat4 type and operations
#include once "arrayList.bi" '' convenience

#include once "camera.bas" '' the camera class definition
#include once "object.bas" '' the object definition
#include once "utility.bi" '' to draw lines an' stuff directly in 3D space
Const as integer scrW = 1366, scrH = 768
Dim Shared nodes As arrayList  = arrayList()
Dim Shared sectors As arrayList Ptr
sectors = New arrayList()

#Include Once "3DGame.bi"
#Include Once "fps.bas"

/'
   3D playground

   intended as a simple tutorial/framework to do 3D stuff without too much complication
   it is a bare-bones implementation, actually. Mostly to test all the math/conventions
   behind the representation of a 3D scene (using the term 'scene' veeeery loosely here)

   conventions used
      RIGHT HANDED coordinate system
      positive rotations are COUNTER-CLOCKWISE
      facets are defined in COUNTER-CLOCKWISE order
      angles are in RADIANS - use radians( angle_in_degrees ) for convenience
'/
'' set a screen mode

function getCpuTick() as double
	dim as double ticks
	asm
		lea edi,[ticks]
		rdtsc
		mov [edi],eax
		mov [edi+4],edx
		fild qword ptr [edi]
		fstp qword ptr [edi]
	end asm
	return ticks
end Function
ScreenRes( scrW, scrH, 32,1 , GFX_ALPHA_PRIMITIVES )

windowTitle( fbVersion & " - 3D Playground" )
InitGFX(0)


/'
   define a projection plane (in this case the screen itself)

   the projection plane is where the image gets projected (duh)
   in the pinhole camera model, the projection plane is actually the
   so called 'near' plane. The near parameter of the camera is really
   the distance of this plane to the origin of coordinates (in the
   case of the camera, its position in the world coordinate system
'/
dim  As rectangle projectionPlane = ( 0.0, 0.0, scrW, scrH )
dim  As rectangle viewPlane = ( scrW-200, scrH-100, scrW, scrH )

/'
   instantiate a camera class
   remember the parameters for the constructor:

   position, in WORLD space
   X axis of the camera's coordinate system
   Y axis of the camera's coordinate system
   Z axis of the camera's coordinate system. If you flip it, you will end with a
      left-handed coordinate system. Not that it matters much, save for the fact that
      all other methods treat the matrix as a right-handed one. If you use lookAt()
      and suddenly the axes get flipped, this is the most probable reason
   near clipping plane Z distance
   far clipping plane Z distance
   projection plane

   a word of advice: the axes of the camera have to be of a certain length, for the
   projection matrix depends on it. See the code below, where the U and V vectors
   get scaled to match the relation of the projection plane. If you fool around with
   them and suddenly the image looks like crap, this is the most probable reason.

   to remember:
      the U vector (the X one) controls the horizontal field of view
      the V vector (the Y one) controls the vertical field of view
      and the Z vector controls the zoom

      of course, they must be perfectly perpendicular to each other, if not the resulting
      image is sheared. Anyway, you don't have to mess with the constructor if you don't
      like and/or understand what it does for the camera to do its job.

   having said all that, play around! See how each function transforms the camera, and
   the effect it has on the image. Look at the implementation to see how it's done, but
   most important of all, have fun!
'/
dim as cameras cam = cameras( _
vec4( 0.0, 0.0, 5.0 ), _
vec4( 1.0, 0.0, 0.0 ), _
vec4( 0.0, 1.0, 0.0 ), _
vec4( 0.0, 0.0, 1.0 ), _
0.1, 10000.0, projectionPlane )
dim as cameras pre_cam = cameras( _
vec4( 0.0, 20.0, 40.0 ), _
vec4( 1.0, 0.0, 0.0 ), _
vec4( 0.0, 1.0, 0.0 ), _
vec4( 0.0, 0.0, 1.0 ), _
0.1, 10000.0, projectionPlane )
Dim as cameras sector_cam = cameras( _
vec4( 0.0,2.0,10.0 ), _
vec4( 1.0, 0.0, 0.0 ), _
vec4( 0.0, 1.0, 0.0 ), _
vec4( 0.0, 0.0, 1.0 ), _
0.1, 10000.0, projectionPlane )
'dim as cameras view_cam = cameras( _
'vec4( 0.0, 5.0, 100.0 ), _
'vec4( 1.0, 0.0, 0.0 ), _
'vec4( 0.0, 1.0, 0.0 ), _
'vec4( 0.0, 0.0, 1.0 ), _
'0.1, 10000.0, viewPlane )

'cam.lookat(vec4(0,0,0,0),vec4(0,1,0,0))
pre_cam.lookat(vec4(0,0,0,0),vec4(0,1,0,0))

'' background color (change it if you don't like it)
dim as uint32 backgroundColor = rgba( 8, 8, 32, 255 )
color( , backgroundColor )





/'----------------------------------------------------------------------------'/
/'                      Test Main for the 3ds-bin lib                         '/
/'----------------------------------------------------------------------------'/

Dim As String filen="test.3ds"
bin3ds=FreeFile

If Open(filen For Binary Access Read As bin3ds) Then End

#Ifdef __DEBUG__
print"Loading 3ds binary file : ";filen
#EndIf
While (ReadPrimaryChunk()=0)
	Print current_chunk,temp_name
Wend
Close #bin3ds
/'----------------------------------------------------------------------------'/

Dim g As galaxy_t Ptr = New galaxy_t(timer)
Dim Shared t As UInteger=Int(0)
Dim Shared s As UInteger
s=Int(Rnd*g->model->vertices->count)
Dim Shared target As vec4 Ptr
Dim Shared target_tmp As vec4 Ptr
target=g->model->vertices->Get(t)
target_tmp=g->model->vertices->Get(t)
Dim cpos1 As vec4 Ptr=g->model->vertices->Get(s)

cam.setpos(*cpos1)
'' (extremely crappy) main loop
Dim play As uinteger=0
Dim up As Integer=-1
Dim As object3d Ptr r
Dim v As vec4=cam.getdir()
Dim v1 As vec4=cam.getdir()
Dim inner_sector As UInteger=0
Dim inner_sector_radi As Double

Do
	Dim cdist As Double
	If play=1 And inner_sector=0 Then
		Dim cpos As vec4=cam.getPos()
		Dim cdir As vec4=cam.getDir()
		cdist=distance3d(target->x,target->y,target->z,cpos.x,cpos.y,cpos.z)
		v =vec4(CurveValue(target->x,v.x,abs(target->x-v.x)*15),CurveValue(target->y,v.y,abs(target->y-v.y)*15),CurveValue(target->z,v.z,abs(target->z-v.z)*15))
		v1 =get_radians_between(target,@cpos)
		cam.lookAt(v,vec4(0,1,0))
		'view_cam.lookat(v,vec4(0,1,0))
		'cam.lookAt(*target_tmp,vec4(0,1,0))

		If ((cdist>0.51) And (inner_sector=0)) Then
			cam.move( normalize( cdir ) * 1.0 * FPS.deltatime/1000)
		ElseIf (cdist<=0.51)Then ' And (r->vertices->count<>t) then
			't+=1
			'target_tmp=r->vertices->Get(t)
			'v=cam.getPos()
			'EndIf
			'If r->vertices->count=t Then
			inner_sector=1
			Dim sy1 As system_t Ptr
			Dim s1 As sector_t Ptr
			Dim v1 As vec4 Ptr
			'Dim n1 As Integer=find_node(@nodes,target)
			s1=sectors->Get(t)
			sy1=s1->systems->Get(0)
			inner_sector_radi=20
			If sy1->star[0].nr_planets>0 Then
				inner_sector_radi=sy1->star[0].planet[sy1->star[0].nr_planets-1].orbit
				If sy1->nr_stars>1 then
					If sy1->star[1].nr_planets>0 Then
						If sy1->star[1].planet[sy1->star[1].nr_planets-1].orbit>inner_sector_radi Then
							inner_sector_radi=sy1->star[1].planet[sy1->star[1].nr_planets-1].orbit
						EndIf
					EndIf
				endif
			EndIf
			sector_cam.setpos(vec4(0,0,inner_sector_radi*AU))
		Else
			'	s=t'Int(Rnd*g->model->vertices->count)
			't=Int(Rnd*g->model->vertices->count)

			'target=g->model->vertices->Get(t)'r->vertices->Get(r->vertices->count-1)
		EndIf
		'If r->vertices->count=t Then
		'	Delete (r)
		'r=NULL
		's=t'Int(Rnd*g->model->vertices->count)
		't=Int(Rnd*g->model->vertices->count)
		''r=get_route(g->model,s,t)
		''t=0'r->vertices->count-1
		'target=g->model->vertices->Get(t)'r->vertices->Get(r->vertices->count-1)
		''target_tmp=r->vertices->Get(t)
		'EndIf
	EndIf



	If play=0 Then
		Dim As vec4 rot=rotateAroundAxis(pre_cam.getpos,vec4(0,1,0),radians(10*FPS.deltatime/1000) )
		pre_cam.setpos(rot )
		Dim y As vec4=pre_cam.getPos
		y.y=50+Sin(Timer*0.25)*25 'up*frameTime*10
		pre_cam.setpos(y)
		pre_cam.lookat(vec4(0,0,0,0),vec4(0,1,0,0))
		if( event->GADGETMESSAGE=FlyButton ) then
			play=1
			Line(0,0)-(scrW,scrH),backgroundColor,BF

			g->pre_render_model(pre_cam)

			g->generate_MST(pre_cam)


			s=0'Int(Rnd*g->model->vertices->count)
			cpos1 =g->model->vertices->Get(s)

			cam.setpos(*cpos1)
			'r=get_route(g->model,s,t)
			t=Int(Rnd*g->model->vertices->count)'r->vertices->count-1
			target=g->model->vertices->Get(t)'r->vertices->Get(r->vertices->count-1)
			'target_tmp=r->vertices->Get(t)
			For i As UInteger=0 To nodes.count-1
				Dim n As node_t Ptr =nodes.get(i)
				generate_sectors(n,timer)
			Next
		EndIf
	endif

	'' render the screen

	screenLock()

	Line(0,0)-(scrW,scrH),backgroundColor,BF
	'Draw String(100,100),Str(degrees(v1.x))&" "&Str(degrees(v1.y))&" "&Str(degrees(v1.z)),&hffffffff
	'Draw String(500,110),Str( g->model->vertices->count),&Hffffffff
	'Draw String(500,120),Str( sectors->count),&Hffffffff
	'Draw String(500,130),Str( nodes.count),&Hffffffff
	Dim sy1 As system_t Ptr
	Dim s1 As sector_t Ptr
	Dim v1 As vec4 Ptr

	If play=0 Then
		g->pre_render_model(pre_cam)
	ElseIf inner_sector=0 And play=1 Then
		g->render_model(cam)

		If Int((Timer*10)Mod 2)=0 Then
			drawPoint3d(cam,g->model->objectSpaceToWorldSpace( *target ),10,&hff00ff00)
		Else
			drawPoint3d(cam,g->model->objectSpaceToWorldSpace( *target ),10,&h336666ff)
		EndIf
		Draw String(10,30),Str(cdist),&hffffffff

		'If r->vertices->count>1 Then

		'	'For i As UInteger=1 To r->vertices->count-1
		'	'	Dim v1 As vec4 Ptr=r->vertices->Get(i-1)
		'	'	Dim v2 As vec4 Ptr=r->vertices->Get(i)
		'	'	drawLine3d(cam,*v1,*v2,&h66ffff00)
		'	'	drawText3d(cam,*v1,Str(i-1),&hffff0000)
		'	'Next
		'EndIf



		'For i As UInteger=0 To sectors->count-1
		'	s1=sectors->Get(i)
		'	Dim n1 As Integer=find_node(@nodes,s1->position)
		'	's1->render_sector(sector_cam,g->model)
		'	s1=sectors->Get(n1)
		'	v1=s1->position
		'	If MouseOver(cam,*v1,s1->size)=TRUE Then
		'		sy1=s1->systems->Get(0)

		'		'drawText3d(cam,*v1,star_class_str(sy1->star[0].star_class),&hffffffff)
		'		Draw String(10,100),sy1->star[0].star_type,&hffffffff
		'		If sy1->star[0].nr_planets>0 Then
		'			'drawText3d(cam,v2,str(sy1->star[0].planet[sy1->star[0].nr_planets-1].orbit),&hffffffff)
		'			draw string(10,120),str(sy1->star[0].planet[sy1->star[0].nr_planets-1].orbit),&hffffffff
		'		EndIf
		'	EndIf
		'Next
	Else
		If inner_sector<>0 And play=1 Then
			'If ((Timer-inner_sector_time) > 5) Then inner_sector=0
			'Dim n1 As Integer=find_node(@nodes,target)
			s1=sectors->Get(t)
			sy1=s1->systems->Get(0)
			Dim v As vec4=sector_cam.getPos()
			Dim d As Double=distance3d(v.x,v.y,v.z,s1->bary_center.x,s1->bary_center.y,s1->bary_center.z)
			Dim cdr As vec4=sector_cam.getDir()

			sector_cam.lookat(s1->bary_center ,vec4(0,1,0))
			If d>0.1 Then
				sector_cam.move(normalize( cdr ) * 1 * FPS.deltatime/1000)
			Else
				inner_sector=0
				s=t'Int(Rnd*g->model->vertices->count)
				t=Int(Rnd*g->model->vertices->count)

				target=g->model->vertices->Get(t)'r->vertices->Get(r->vertices->count-1)

			EndIf
			s1->render_sector(sector_cam,g->model)
			Draw String(10,30),Str(d),&hffffffff
			Draw String(10,100),sy1->star[0].star_type,&hffffffff
			Draw String(10,110),Str(sy1->star[0].luminousity),&hffffffff
			Draw String(10,120),Str(sy1->star[0].temperature),&hffffffff
			Draw String(10,130),Str(sy1->star[0].radius),&hffffffff
			Draw String(10,140),Str(sy1->star[0].mass),&hffffffff
			Draw String(10,150),Str(sy1->star[0].orbit)&" "&Str(sy1->star[0].nr_planets),&hffffffff
			If sy1->nr_stars>1 Then
				Draw String(10,200),sy1->star[1].star_type,&hffffffff
				Draw String(10,210),Str(sy1->star[1].luminousity),&hffffffff
				Draw String(10,220),Str(sy1->star[1].temperature),&hffffffff
				Draw String(10,230),Str(sy1->star[1].radius),&hffffffff
				Draw String(10,240),Str(sy1->star[1].mass),&hffffffff
				Draw String(10,250),Str(sy1->star[1].orbit)&" "&Str(sy1->star[1].nr_planets),&hffffffff
			EndIf
		endif
	EndIf
	'Print t,Log(1+0.1),Log(1+360*680000),Log(1+50*149600000)

	If play=0 Then
		Line (0,scrH-150-20)-(420,scrH),&hff000000,BF
		Line (0,scrH-150-20)-(420,scrH),&hffffffff,B
		GadgetOn(ScrollBulgeSize)
		GadgetOn(ScrollBulgeHeight)
		GadgetOn(ScrollBulgeOffset)
		GadgetOn(ScrollSpiralK)
		GadgetOn(ScrollSpiralP)
		GadgetOn(ScrollSpiralArms)
		GadgetOn(FlyButton)
		Draw String (10,scrH-150+04),"Bulge Size:",&hffffffff
		Draw String (360,scrH-150+04),Str(GetScrollBarVal(ScrollBulgeSize)),&hffffffff
		Draw String (10,scrH-150+44),"Bulge Height:",&hffffffff
		Draw String (360,scrH-150+44),Str(GetScrollBarVal(ScrollBulgeHeight)),&hffffffff
		Draw String (10,scrH-150+24),"Bulge Offset:",&hffffffff
		Draw String (360,scrH-150+24),Str(GetScrollBarVal(ScrollBulgeOffset)),&hffffffff
		Draw String (10,scrH-150+64),"Spiral K:",&hffffffff
		Draw String (360,scrH-150+64),Str(GetScrollBarVal(ScrollSpiralK)),&hffffffff
		Draw String (10,scrH-150+84),"Spiral P:",&hffffffff
		Draw String (360,scrH-150+84),Str(GetScrollBarVal(ScrollSpiralP)),&hffffffff
		Draw String (10,scrH-150+104),"Nr. Spiral Arms:",&hffffffff
		Draw String (10,scrH-150+124),"Nr. Systems: " & Str(g->model->vertices->count),&hffffffff
		Draw String (360,scrH-150+104),Str(GetScrollBarVal(ScrollSpiralArms)),&hffffffff
	Else
		GadgetOff(ScrollBulgeSize)
		GadgetOff(ScrollBulgeHeight)
		GadgetOff(ScrollBulgeOffset)
		GadgetOff(ScrollSpiralK)
		GadgetOff(ScrollSpiralP)
		GadgetOff(ScrollSpiralArms)
		GadgetOff(FlyButton)
	EndIf
	FPS.update()
	Draw String (10,0),"FT: " & format(FPS.deltatime,"0.0000") & " ms",&hffffffff
	Draw String (10,10),"FPS: " & format(FPS.FC,"0.0000"),&hffffffff

	screenUnlock()
	'ScreenSync

	event->xSleep(0)
	if event->GADGETMESSAGE then
		select case event->GADGETMESSAGE
			case ScrollBulgeSize
				g->regenerate()
			case ScrollBulgeHeight
				g->regenerate()
			Case ScrollBulgeOffset
				g->regenerate()
			Case ScrollSpiralK
				g->regenerate()
			Case ScrollSpiralP
				g->regenerate()
			Case ScrollSpiralArms
				g->regenerate()
				'play=Not(play)
				' print "Scrollbereich:" & minimum & "-" & maximum & " @ Pagesize " & pagesize & ": " & GetScrollBarVal(Scroll)
		end select
	end If
	If KEY="+" And t<sectors->count-1 Then t+=1

	If KEY="-" And t>0 Then t-=1
	'Sleep 1
loop until ( MultiKey( fb.sc_escape )) Or (event->EXITEVENT )'' loop until esc is pressed or x pressed

Delete event
If sectors->count()>0 Then
	For i As UInteger=0 To sectors->count()-1
		sectors->removeLast()
	Next
	Delete sectors
EndIf