#Include once "platform.bi" '' platform specific definitions
#include once "core.bi" '' core functions and defines
#include once "math.bi" '' math functions and constants
#include once "vec4.bi" '' vec4 type and operations
#include once "mat4.bi" '' mat4 type and operations

/'
   this serves to define a projection plane
   
   in the code, x and y coordinates are not used. It only serves to replace
   the IRenderContext interface
'/
type rectangle
   as double x
   as double y
   as double width
   as double height
end type

/'
   simple camera class
   
   responsibility: represents a pinhole-model camera. Can be used to navigate a scene FPS style
      very useful for debugging purposes (especially geometry code)
   mutability: mutable. Use with caution or solely for debugging. Just don't try to write Doom 5 with this.
   
   10/10/2017: refactored the code because it's a mess.       
'/

class cameras
  public:
      declare constructor( _
         byval position as vec4, _
         byval x as vec4, _
         byval y as vec4, _
         byval z as vec4, _
         byval near as double = 1.0, _
         byval far as double = 10000.0, _
         byval projPlane as rectangle )
        
        '' lots of crappy functions to play with
         declare function getU() as vec4
         declare function getV() as vec4
         declare function getDir() as vec4
         declare function getPos() as vec4         
         declare sub move( byval offset as vec4 )
         declare sub rotate( byval axis as vec4, byval angle as double )
         declare function getRatioUV() as double
         declare function getRatioVU() as double
         declare sub setRatioUV( byval ratio as double )
         declare sub setRatioVU( byval ratio as double)
         declare function getZoomU() as double
         declare function getZoomV() as double
         declare sub setZoomU( byval a as double )
         declare sub setZoomV( byval a as double )
         declare sub zoom( byval a as double )
         declare sub zoomU( byval a as double )
         declare sub zoomV( byval a as double )
         declare function getFOVV() as double
         declare function getFOVU() as double
         declare sub setFOVU( byval angle as double )
         declare sub setFOVV( byval angle as double )
         declare function getYaw() as double
         declare function getPitch() as double
         declare function getRoll() as double
         declare sub setYaw( byval angle as double )
         declare sub setPitch( byval angle as double )
         declare sub setRoll( byval angle as double )
         declare sub yawAbsolute( byval angle as double )
         declare sub yaw( byval angle as double )
         declare sub pitch( byval angle as double )
         declare sub roll( byval angle as double )
         declare function getCameraMatrix() as mat4
         declare function getInverseCameraMatrix() as mat4
         declare function transform( byval v as vec4 ) as vec4
         declare function orthogonal( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
         declare function perspective( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
         declare function projectOnScreen( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
         declare function projectOnScreenOrtho( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
         declare function getScale() as double
         declare sub setScale( byval dirLength as double )
         declare sub scale( byval factor as double )
         declare sub resetSkewU()
         declare sub resetSkewV()
         declare function getDistance( byval pnt as vec4 ) as double
         declare sub setDistance( byval pnt as vec4, byval dist as double )
         declare sub setLookDirection( byval newDir as vec4 )
         declare sub lookAt( byval target as vec4 )
         declare sub lookAt( byval target as vec4, byval up as vec4 )
         declare sub setU( byval newU as vec4 )
         declare sub setV( byval newV as vec4 )
         declare sub setDir( byval newDir as vec4 )
         declare sub setPos( byval newPos as vec4 )
        declare function nearClip() as double
        declare function farClip() as double
        declare function projectionPlane() as rectangle
        
  private:
      declare sub generateMatrix()
    
     '' members
     as vec4 m_pos
     as vec4 m_u
     as vec4 m_v
     as vec4 m_dir
      as mat4 m_position
      
     as mat4 m_camMatrix
     as mat4 m_invCamMatrix
   
     as double m_nearClip
     as double m_farClip
   
     as double m_fov
     as rectangle m_projPlane
end class

constructor cameras( _
   byval position as vec4, _
   byval x as vec4, _
   byval y as vec4, _
   byval z as vec4, _
   byval near as double = 1.0, _
   byval far as double = 10000.0, _
   byval projPlane as rectangle )
   
   m_pos = position
   m_u = x
   m_v = y
   m_dir = z
   m_nearClip = near
   m_farClip = far
   m_position = matrices.translation( position )
   m_projPlane = projPlane
   
   /'
      adjust the aspect ratio of the camera to match the
      projection plane resolution
      
      sorry for this crap, I had to reimplement it this way to shorten
      the code
   '/
   with m_projPlane
      if( .width < .height ) then
         setRatioUV( .width / .height )
      else
         setRatioVU( .height / .width )
      end if
   end with
   
   generateMatrix()
end constructor

sub cameras.generateMatrix()
  '' generates the camera's rotation matrix
  with m_camMatrix
      .a = m_u.x : .b = m_v.x : .c = m_dir.x : .d = 0.0
      .e = m_u.y : .f = m_v.y : .g = m_dir.y : .h = 0.0
      .i = m_u.z : .j = m_v.z : .k = m_dir.z : .l = 0.0
      .m = m_u.w : .n = m_v.w : .o = m_dir.w : .p = 1.0
  end with
   
   '' and calculates the inverse immediately
   m_invCamMatrix = inverse( m_camMatrix )
end sub

sub cameras.move( byval offset as vec4 )
   m_pos = m_pos + offset   
end sub

sub cameras.rotate( byval axis as vec4, byval angle as double )
   '' rotates the camera
   m_u = rotateAroundAxis( m_u, axis, angle )
   m_v = rotateAroundAxis( m_v, axis, angle )
   m_dir = rotateAroundAxis( m_dir, axis, angle )
      
   generateMatrix()
end sub

function cameras.projectionPlane() as rectangle
   return( m_projPlane )
end function

function cameras.getU() as vec4
   return( m_u )
end function

function cameras.getV() as vec4
   return( m_v )
end function

function cameras.getDir() as vec4
   return( m_dir )
end function

function cameras.getPos() as vec4
   return( m_pos )
end function

function cameras.getRatioUV() as double
   return( m_u.length() / m_v.length() )
end function

function cameras.getRatioVU() as double
   return( m_v.length() / m_u.length() )
end function

sub cameras.setRatioUV( byval ratio as double )
   m_v.normalize()
   m_v = m_v * m_u.length() / ratio
   
   generateMatrix()
end sub

sub cameras.setRatioVU(byval ratio as double)
   m_u.normalize()
   m_u = m_u * m_v.length() / ratio
   
   generateMatrix()
end sub
 
function cameras.getZoomU() as double
   return( m_dir.length() / m_u.length() )
end function

function cameras.getZoomV() as double
   return( m_dir.length() / m_v.length() )
end function

sub cameras.setZoomU( byval a as double )
   m_u = m_u / ( a / getZoomU() )
   
   generateMatrix()
end sub

sub cameras.setZoomV( byval a as double )
   m_v = m_v / ( a / getZoomV() )
   
   generateMatrix()
end sub

sub cameras.zoom( byval a as double )
   zoomU( a )
   zoomV( a )
end sub

sub cameras.zoomU( byval a as double )
   m_u = m_u / a
   
   generateMatrix()
end sub

sub cameras.zoomV( byval a as double )
   m_v = m_v / a
   
   generateMatrix()
end sub

function cameras.getFOVV() as double
   return( 2.0 * atan2( m_v.length(), m_dir.length() ) )
end function

function cameras.getFOVU() as double
   return( 2.0 * atan2( m_u.Length(), m_dir.length() ) )
end function

sub cameras.setFOVU( byval angle as double )
   setzoomU( 1.0 / tan( angle / 2.0 ) )
end sub

sub cameras.setFOVV( byval angle as double )
   setzoomV( 1.0 / tan( angle / 2.0 ) )
end sub

function cameras.getYaw() as double
   return( atan2( m_dir.x, m_dir.z ) )
end function

function cameras.getPitch() as double
   return( atan2( m_dir.y, sqr( m_dir.x * m_dir.x + m_dir.z * m_dir.z ) ) )
end function

function cameras.getRoll() as double
  return( vectorAngle( cross( vec4( 0.0, 1.0, 0.0 ), m_dir ), cross( m_v, m_dir ) ) )
end function

sub cameras.setYaw( byval angle as double )
  '' to change yaw, you have to rotate around the "up" axis of the WORLD = the y axis
  dim currentAngle as double = getyaw()
 
  rotate( vec4( 0.0, 1.0, 0.0 ), -angle + currentAngle )
end sub

sub cameras.setPitch( byval angle as double )
   dim currentAngle as double = getPitch()
   
   rotate( m_u, angle - currentAngle )
end sub

sub cameras.setRoll( byval angle as double )
   dim currentAngle as double = getRoll()
   
   rotate( m_dir, angle - currentAngle )
end sub

sub cameras.yawAbsolute( byval angle as double )
   rotate( vec4( 0.0, 1.0, 0.0 ), angle )
end sub

sub cameras.yaw( byval angle as double )
   rotate( m_v, angle)
end sub

sub cameras.pitch(byval angle as double)
   rotate( m_u, angle )
end sub

sub cameras.roll( byval angle as double )
   rotate( m_dir, angle )
end sub

function cameras.getCameraMatrix() as mat4
   return( m_camMatrix )
end function

function cameras.getInverseCameraMatrix() as mat4
   return( m_invCamMatrix )
end function

function camera.transform( byval v as vec4 ) as vec4
   return( m_invCamMatrix * v )
end function

function cameras.orthogonal( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
   dim as mat4 PM
   dim as double ratio
   
   if( plane.width < plane.height ) then
      ratio = plane.height
   else
      ratio = plane.width
   end if
   
   with PM
      .a = ( 2 / ratio ) * 7 : .b = 0.0 : .c = 0.0 : .d = 0.0
      .e = 0.0 : .f = -( 2 / ratio ) * 7 : .g = 0.0 : .h = 0.0
      .i = 0.0 : .j = 0.0 : .k = ( -2.0 / ( m_farClip - m_nearClip ) ) * 7 : .l = ( ( m_farClip + m_nearClip ) / ( m_farClip - m_nearClip ) )
      .m = 0.0 : .n = 0.0 : .o = 0.0 : .p = 1.0
   end with
   
   Dim as vec4 b
   b = PM * pnt
   
   dim as double px, py
   
   px = ( b.x * plane.width ) + plane.width / 2
   py = ( b.y * plane.height ) + plane.height / 2
   
   if( px >= 0 and px <= plane.width - 1 and py >= 0 and py <= plane.height - 1 ) then
       x = px
       y = py
       z = b.z
   
       return( -1 )
   else
       x = px
       y = py
       z = b.z
      
       return( 0 )
   end if
end function

function cameras.perspective( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer   
   '' the current matrix
   dim as double fovX = radians( 90 )
   dim as double fovY = radians( 90 )
   dim as double fovTanX = 1 / tan( fovX / 2 )
   dim as double fovTanY = 1 / tan( fovY / 2 )
   
   '' projection matrix (assumes a right-handed coord system)
  dim as mat4 PM
 
  with PM
      .a = fovTanX : .b = 0.0 : .c = 0.0 : .d = 0.0
      .e = 0.0 : .f = fovTanY : .g = 0.0 : .h = 0.0
      .i = 0.0 : .j = 0.0 : .k = -( ( farClip + nearClip ) / ( farClip - nearClip ) ) : .l = -( ( 2 * ( farClip * nearClip ) ) / ( farClip - nearClip ) )
      .m = 0.0 : .n = 0.0 : .o = -1.0 : .p = 0.0
  end with
   
   '' project the point
   dim as vec4 b
   b = PM * pnt
   
   dim as double px
   dim as double py
   
   px = ( b.x * plane.width ) / ( 1.0 * b.w ) + plane.width / 2
   py = ( b.y * plane.height ) / ( 1.0 * b.w ) + plane.height / 2   
   
   if( px >= 0 and px <= plane.width - 1 and py >= 0 and py <= plane.height - 1 and b.z > m_nearClip ) then
       x = px
       y = py
       z = b.z
   
       return( -1 )
   else
       x = px
       y = py
       z = b.z
   
       return( 0 )
   end if
end function

function cameras.projectOnScreen( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
   '' first transformation: translation
   dim as mat4 TM = matrices.translation( -m_pos )
   dim as vec4 a = vec4( TM * pnt )
   
   '' second transformation: rotation
   '' transform() multiplies the vector with the INVERSE camera matrix
   '' to bring the point from camera space to world space
   dim as vec4 b = vec4( transform( a ) )
   
   '' Third transformation: projection
   return( perspective( b, plane, x, y, z ) )
end function

function cameras.projectOnScreenOrtho( byval pnt as vec4, byval plane as rectangle, byref x as double, byref y as double, byref z as double ) as integer
   dim as mat4 TM
   
   '' first transformation: translation
   with TM
       .a = 1.0 : .b = 0.0 : .c = 0.0 : .d = -m_pos.x
       .e = 0.0 : .f = 1.0 : .g = 0.0 : .h = -m_pos.y
       .i = 0.0 : .j = 0.0 : .k = 1.0 : .l = -m_pos.z
       .m = 0.0 : .n = 0.0 : .o = 0.0 : .p = 1.0
   end with
   
   dim as vec4 a = vec4( TM * pnt )
   
   '' transform() multiplies the vector with the INVERSE camera matrix
   '' to bring the point from camera space to world space
   dim as vec4 b = vec4( transform( a ) )
   
   '' third transformation: projection
   return( orthogonal( b, plane, x, y, z ) )
end function

function cameras.getScale() as double
   return( m_dir.length )
end function

sub cameras.setScale( byval dirLength as double )
   scale( m_dir.length() / dirLength )
end sub

sub cameras.scale( byval factor as double )
   m_dir = m_dir * factor
   m_u = m_u * factor
   m_v = m_v * factor
   
   generateMatrix()
end sub

sub cameras.resetSkewU()
   dim as double oldzoomU = getzoomU()
   dim as double oldzoomV = getzoomV()
   
   m_u = cross( m_dir, m_v )
   m_v = cross( m_dir, -m_u )
   
   setzoomU( oldzoomU )
   setzoomV( oldzoomV )
end sub

sub cameras.resetSkewV()
   dim as double oldzoomU = getzoomU()
   dim as double oldzoomV = getzoomV()
   
   m_v = cross( m_dir, m_u )
   m_u = cross( m_dir, -m_v )
   
   setzoomU( oldzoomU )
   setzoomV( oldzoomV )
end sub

sub cameras.setLookDirection( byval newDir as vec4 )
   dim as vec4 axis = vec4( ( cross( m_dir, newDir ) ) )
   
   if axis.length() = 0 then
      exit sub
   end if
   
   dim as double angle = vectorAngle( m_dir, newDir )
   
   if angle <> 0 then
      rotate( axis, angle )
   end if
end sub

sub cameras.lookAt( byval target as vec4, byval up as vec4 )
   /'
      preserve the lengths of the camera vectors
      
      every time one messes with the camera vectors directly, it has to
      remember to preserve their length, if you do not do this, the projection
      gets distorted, as these vectors control different parameters of the
      camera model
         REMEMBER
            u controls the horizontal FOV
            v controls the vertical FOV
            dir controls the focal zoom
   '/
   dim as double dl = m_dir.length()
   dim as double ul = m_u.length()
   dim as double vl = m_v.length()
      
   m_dir = ( normalize( target - m_pos ) * dl )
   m_U = ( normalize( cross( up, m_dir ) ) * ul )
   m_V = ( normalize( cross( m_dir, m_u ) ) * vl )
   
   generateMatrix()
end sub

sub cameras.lookAt( byval target as vec4 )
   /'
      preserve the lengths of the camera vectors
      
      every time one messes with the camera vectors directly, it has to
      remember to preserve their length, if you do not do this, the projection
      gets distorted, as these vectors control different parameters of the
      camera model
         REMEMBER
            u controls the horizontal FOV
            v controls the vertical FOV
            dir controls the focal zoom
   '/
   dim as double dl = m_dir.length()
   dim as double ul = m_u.length()
   dim as double vl = m_v.length()
   
   '' compute the forward vector
   dim as vec4 forward = vec4( normalize( target - m_pos ) )
   
   '' compute temporal up vector based on the forward vector
   '' watch out when look up/down at 90 degree
   '' for example, forward vector is on the Y axis
   dim as vec4 up
   
   if( abs( forward.x ) < epsilon and abs( forward.z ) < epsilon ) then
      if( forward.y > 0 ) then
       '' forward vector is pointing +Y axis
         up = vec4( 0.0, 0.0, -1.0 )
      else
         '' forward vector is pointing -Y axis
         up = vec4( 0.0, 0.0, 1.0 )
      end if
   else
      '' in general, up vector is straight up
      up = vec4( 0.0, 1.0, 0.0 )
   end if
      
   '' compute the left vector
   dim as vec4 leftV = vec4( normalize( cross( up, forward ) ) )
   
   '' re-calculate the orthonormal up vector
   up = normalize( cross( forward, leftV ) )
   
   m_dir = forward * dl
   m_u = leftV * ul
   m_v = up * vl

   generateMatrix()
end sub

sub cameras.setU( byval newU as vec4 )
   m_u = newU
   
   generateMatrix()
end sub

sub cameras.setV( byval newV as vec4 )
   m_v = newV
   
   generateMatrix()
end sub

sub cameras.setDir( byval newDir as vec4 )
   m_dir = newDir
   
   generateMatrix()
end sub

sub cameras.setPos( byval newPos as vec4 )
   m_pos = newPos
end sub

function cameras.nearClip() as double
   return( m_nearClip )
end function

function cameras.farClip() as double
   return( m_farClip )
end function

function cameras.getDistance( byval pnt as vec4 ) as double
   return( distance( m_pos, pnt ) )
end function

public sub cameras.setDistance( byval pnt as vec4, byval dist as double )
  dim currentDist as double = distance( m_pos, pnt )
  move( currentDist * normalize( pnt - m_pos ) )
end sub
