'' needed headers
#include once "platform.bi"
#include once "core.bi"
#include once "math.bi"
#include once "vec4.bi"
#include once "mat4.bi"
#include once "arrayList.bi"

type edge
   public:
      declare constructor( byval v1 as vec4 ptr, byval v2 as vec4 ptr )
      
      declare function vertex1() as vec4 ptr
      declare function vertex2() as vec4 ptr
      
   private:
      as vec4 ptr m_v1
      as vec4 ptr m_v2
end type

constructor edge( byval v1 as vec4 ptr, byval v2 as vec4 ptr )
   m_v1 = v1
   m_v2 = v2
end constructor

function edge.vertex1() as vec4 ptr
   return( m_v1 )
end function

function edge.vertex2() as vec4 ptr
   return( m_v2 )
end function

class object3D
   public:
      declare constructor()
      
      declare constructor( _
         byval position as vec4 = vec4( 0.0, 0.0, 0.0 ), _
         byval xAxis as vec4, _
         byval yAxis as vec4, _
         byval zAxis as vec4 )
      
      declare destructor()
      
      declare function transform( byval v as vec4 ) as vec4
      declare function objectSpaceToWorldSpace( byval pnt as vec4 ) as vec4
      declare function getObjectMatrix() as mat4
      declare function getInverseObjectMatrix() as mat4
      declare sub move( byval offset as vec4 )
      declare sub rotate( byval axis as vec4, byval angle as double )
      declare function getyaw() as double
      declare function getPitch() as double
      declare function getRoll() as double
      declare sub setyaw( byval angle as double )
      declare sub setPitch( byval angle as double )
      declare sub setRoll( byval angle as double )
      declare sub yawAbsolute( byval angle as double )
      declare sub yaw( byval angle as double )
      declare sub pitch( byval angle as double )
      declare sub roll( byval angle as double )
      declare function getDistance( byval pnt as vec4 ) as double
      declare sub setDistance( byval pnt as vec4, byval dist as double )
      declare sub setLookDirection( byval newDir as vec4 )
      declare sub lookAt( byval lookAtMe as vec4 )
      declare sub setU( byval newU as vec4 )
      declare sub setV( byval newV as vec4 )
      declare sub setDir( byval newDir as vec4 )
      declare sub setPos( byval newPos as vec4 )
      declare function getU() as vec4
      declare function getV() as vec4
      declare function getDir() as vec4
      declare function getPos() as vec4
      
      '' these two functions are wrong on so many levels that isn't even funny
      '' made it this way to simplify a little
      declare function vertices() as arrayList ptr
      declare function edges() as arrayList ptr
      
   private:
      declare sub generateMatrix()
      
      as vec4 m_pos = any
      as vec4 m_u = any
      as vec4 m_v = any
      as vec4 m_dir = Any
      
      as mat4 m_objMatrix = any
      as mat4 m_invObjMatrix = any
      
      as arrayList m_vertices
	   As arrayList m_col 
      as arrayList m_edges
End class

constructor object3d()
   '' construct an object with the default axes
   m_pos = vec4( 0.0, 0.0, 0.0 )
   m_u = vec4( 1.0, 0.0, 0.0 )
   m_v = vec4( 0.0, 1.0, 0.0 )
   m_dir = vec4( 0.0, 0.0, 1.0 )
   
   generateMatrix()
end constructor

constructor object3d( _
   byval position as vec4 = vec4( 0.0, 0.0, 0.0 ), _
   byval xAxis as vec4, _
   byval yAxis as vec4, _
   byval zAxis as vec4 )
   
   m_pos = position
   m_u = xAxis
   m_v = yAxis
   m_dir = zAxis
   
   generateMatrix()
end constructor

destructor object3d()
   '' if there are vertices on the object, release them
   dim as vec4 ptr v
   
   if( m_vertices.count > 0 ) then
      for i as integer = 0 to m_vertices.count - 1
         v = m_vertices.get( i )
         delete( v )
      next
   end if
   
   '' same for the edges
   dim as edge ptr e
   
   if( m_edges.count > 0 ) then
      for i as integer = 0 to m_edges.count - 1
         e = m_edges.get( i )
         delete( e )
      next
   end if
end destructor

function object3d.vertices() as arrayList ptr
   return( @m_vertices )
end function

function object3d.edges() as arrayList ptr
   return( @m_edges )
end function

function object3d.getObjectMatrix() as mat4
   return( m_objMatrix )
end function

function object3d.getInverseObjectMatrix() as mat4
   return( m_invObjMatrix )
end function

sub object3d.generateMatrix()
  '' generates the object rotation matrix
  m_objMatrix = mat4( _
      m_u.x , m_v.x , m_dir.x , 0.0, _
      m_u.y , m_v.y , m_dir.y , 0.0, _
      m_u.z , m_v.z , m_dir.z , 0.0, _
      m_u.w , m_v.w , m_dir.w , 1.0 )
   
   '' and compute the inverse immediately
   m_invObjMatrix = inverse( m_objMatrix )
end sub

sub object3d.move( byval offset as vec4 )
   m_pos += offset
end sub

sub object3d.rotate( byval axis as vec4, byval angle as double )
   '' rotates the object around an arbitrary axis
   m_u = rotateAroundAxis( m_u, axis, angle)
   m_v = rotateAroundAxis( m_v, axis, angle)
   m_dir = rotateAroundAxis( m_dir, axis, angle)
   
   generateMatrix()
end sub

function object3d.getU() as vec4
   return( m_u )
end function

function object3d.getV() as vec4
   return( m_v )
end function

function object3d.getDir() as vec4
   return( m_dir )
end function

function object3d.getPos() as vec4
   return( m_pos )
end function

function object3d.getyaw() as double
   return( atan2( m_dir.x, m_dir.z ) )
end function

function object3d.getPitch() as double
   return( atan2( m_dir.y, sqr( m_dir.x * m_dir.x + m_dir.z * m_dir.z ) ) )
end function

function object3d.getRoll() as double
  return( vectorAngle( cross( vec4( 0.0, 1.0, 0.0 ), m_dir ), cross( m_v, m_dir ) ) )
end function

sub object3d.setyaw( byval angle as double )
  '' to change yaw, you have to rotate around the "up" axis of the WORLD = the y axis
  dim as double currentAngle = getyaw()
  rotate( vec4( 0.0, 1.0, 0.0 ), -angle + currentAngle )
end sub

sub object3d.setPitch( byval angle as double )
  dim as double currentAngle = getPitch()
  rotate( m_u, angle - currentAngle )
end sub

sub object3d.setRoll( byval angle as double )
   dim as double currentAngle = getRoll()
   rotate( m_dir, angle - currentAngle )
end sub

sub object3d.yawAbsolute( byval angle as double )
   rotate( vec4( 0.0, 1.0, 0.0 ), angle )
end sub

sub object3d.yaw( byval angle as double )
   rotate( m_v, angle )
end sub

sub object3d.pitch( byval angle as double )
   rotate( m_u, angle )
end sub

sub object3d.roll( byval angle as double )
   rotate( m_dir, angle )
end sub

function object3d.getDistance( byval pnt as vec4 ) as double
   return( distance( m_pos, pnt ) )
end function

sub object3d.setDistance( byval pnt as vec4, byval dist as double )
   dim as double currentDist = distance( m_pos, pnt )
   move( currentDist * normalize( pnt - pos ) )
end sub

sub object3d.setLookDirection( byval newDir as vec4 )
   dim axis as vec4 = vec4( ( cross( m_dir, newDir ) ) )
   
   if axis.length() = 0 then
      exit sub
   end if
   
   dim as double angle = vectorAngle( m_dir, newDir )
   
   if( angle <> 0 ) then
      rotate( axis, angle )
   end if
end sub

sub object3d.lookAt( byval lookAtMe as vec4 )
   dim as double dl = m_dir.length()
   dim as double ul = m_u.length()
   dim as double vl = m_v.length()
   
   dim as vec4 w = vec4( lookAtMe - m_pos )
      
   setDir( Normalize( w ) * dl )
   setU( normalize( cross( vec4( 0.0, 1.0, 0.0 ), m_dir ) ) * ul )
   setV( normalize( cross( m_dir, m_u ) ) * vl )
end sub

sub object3d.setU( byval newU as vec4 )
   m_u = newU
   generateMatrix()
end sub

sub object3d.setV( byval newV as vec4 )
   m_v = newV
   generateMatrix()
end sub

sub object3d.setDir( byval newDir as vec4 )
   m_dir = newDir
   generateMatrix()
end sub

sub object3d.setPos( byval newPos as vec4 )
   m_pos = newPos
end sub

function object3d.transform( byval v as vec4 ) as vec4
   return( m_objMatrix * v )
end function

function object3d.objectSpaceToWorldSpace( byval pnt as vec4 ) as vec4
  dim as vec4 b = vec4( transform( pnt ) )
  dim as vec4 a = vec4( b + m_pos )
 
  return( a )
end function
