#ifndef __mat3__   
   #define __mat3__

   #include once "vec2h.bi"
   #include once "math.bi"
   /'
                             | a b c |
      3x3 Matrix type      | d e f |
                            | g h i |
      
      note that this type also define operators to work with
      2D homogeneous vectors
      
      08/28/2017: fixed typo in multiplication of two matrices
   '/   
   type mat3
      public:
         as double a, b, c
         as double d, e, f
         as double g, h, i
      
      declare constructor()
      declare constructor(_
         byval na as double, byval nb as double, byval nc as double, _
         byval nd as double, byval ne as double, byval nf as double, _
         byval ng as double, byval nh as double, byval ni as double )
      
      declare constructor( byref rhs as mat3 )
      declare operator let( byref rhs as mat3 )
      
      declare sub transpose()
      declare function determinant() as double
      declare sub inverse()
      declare sub identity()
      
      declare operator cast() as string
   end type
         
   constructor mat3() export
      '' The default constructor creates an identity matrix, instead of a matrix of zeros
      a = 1.0: b = 0.0: c = 0.0
      d = 0.0: e = 1.0: f = 0.0
      g = 0.0: h = 0.0: i = 1.0
   end constructor
   
   constructor mat3(   byval na as double, byval nb as double, byval nc as double, _
                                    byval nd as double, byval ne as double, byval nf as double, _
                                    byval ng as double, byval nh as double, byval ni as double ) export
                                    
      a = na: b = nb: c = nc
      d = nd: e = ne: f = nf
      g = ng: h = nh: i = ni
   end constructor
   
   constructor mat3( byref rhs as mat3 ) export
      '' copy constructor
      a = rhs.a: b = rhs.b: c = rhs.c
      d = rhs.d: e = rhs.e: f = rhs.f
      g = rhs.g: h = rhs.h: i = rhs.i   
   end constructor
   
   operator mat3.let( byref rhs as mat3 ) export
      '' assignment constructor
      a = rhs.a: b = rhs.b: c = rhs.c
      d = rhs.d: e = rhs.e: f = rhs.f
      g = rhs.g: h = rhs.h: i = rhs.i
   end operator

   sub mat3.transpose() export
      /'
         transpose:

         | a b c |T    | a d g |
         | d e f |  =  | b e h |
         | g h i |     | c f i |
      '/
    b += d: d = b - d: b -= d '' swap b and d
    c += g: g = c - g: c -= g '' swap c and g
    f += h: h = f - h: f -= h '' swap f and h
   end sub
   
   function transpose( byval A as mat3 ) as mat3 export
      A.transpose()
      
      return( A )
   end function
   
   function mat3.determinant() as double export
    '' computes the determinant of the matrix
    dim as double det = _
         ( a * e * i _
         + b * f * g _
         + d * h * c _
         - g * e * c _
         - d * b * i _
         - h * f * a )
      
      '' this is, of course, not matematically correct but it saves
      '' some comprobations when you're calculating the inverse of
      '' the matrix (and to avoid dividing by zero)
      if det = 0 then det = 1
      
      return( det )
   end function
   
   function determinant( byval A as mat3 ) as double export
      return( A.determinant() )
   end function
   
   sub mat3.inverse() export
      '' inverse of a 3x3 matrix
      '' the inverse is the adjoint divided through the determinant
      dim as double det = this.determinant()
      
      '' if the determinant is 0, the matrix has no inverse
      if det > 0 then
         dim as double rDet = 1 / det '' The reciprocal of the determinant
         
         '' included in these calculations: minor, cofactor (changed signs), transpose
         '' and division through determinant (expressed here by multiplying by the reciprocal)
         dim as mat3 INV = mat3( _
            (  e * i - h * f ) * rDet, ( -b * i + h * c ) * rDet, (  b * f - e * c ) * rDet, _
            ( -d * i + g * f ) * rDet, (  a * i - g * c ) * rDet, ( -a * f + d * c ) * rDet, _
            (  d * h - g * e ) * rDet, ( -a * h + g * b ) * rDet, (  a * e - d * b ) * rDet )
         
         '' set this matrix to the computed inverse
         this = INV
      end if
   end sub
   
   function inverse( byval A as mat3 ) as mat3 export
      A.inverse()
      
      return( A )
   end function
   
   sub mat3.identity() export
      '' make the matrix an identity matrix
      a = 1.0: b = 0.0: c = 0.0
      d = 0.0: e = 1.0: f = 0.0
      g = 0.0: h = 0.0: i = 1.0
   end sub
   
   operator mat3.cast() as string export
      '' this is only to obtain a human readable representation of the matrix
      return( _
         "[" & trim( str( a ) ) & "][" & trim( str( b ) ) & "][" & trim( str( c ) ) & "]" & chr( 10 ) & chr( 13 ) & _
         "[" & trim( str( d ) ) & "][" & trim( str( e ) ) & "][" & trim( str( f ) ) & "]" & chr( 10 ) & chr( 13 ) & _
         "[" & trim( str( g ) ) & "][" & trim( str( h ) ) & "][" & trim( str( i ) ) & "]" )
   end operator
   
   operator + ( byref A as mat3, byref B as mat3 ) as mat3 export
    '' adds two matrices
    return( mat3( _
         A.a + B.a, A.b + B.b, A.c + B.c, _
         A.d + B.d, A.e + B.e, A.f + B.f, _
         A.g + B.g, A.h + B.h, A.i + B.i ) )
   end operator
   
   operator - ( byref A as mat3, byref B as mat3 ) as mat3 export
      '' substracts two matrices
    return( mat3( _
       A.a - B.a, A.b - B.b, A.c - B.c, _
       A.d - B.d, A.e - B.e, A.f - B.f, _
       A.g - B.g, A.h - B.h, A.i - B.i ) )
   end operator
   
   operator * ( byref A as mat3, byref s as double ) as mat3 export
      '' multiply a matrix with a scalar
    return( mat3( _
         A.a * s, A.b * s, A.c * s, _
         A.d * s, A.e * s, A.f * s, _
         A.g * s, A.h * s, A.i * s ) )
   end operator
   
   operator * ( byref s as double, byref A as mat3 ) as mat3 export
      '' multiply a matrix with a scalar   
    return( mat3( _
       A.a * s, A.b * s, A.c * s, _
       A.d * s, A.e * s, A.f * s, _
       A.g * s, A.h * s, A.i * s ) )
   end operator
   
   operator * ( byref A as mat3, byref B as mat3 ) as mat3 export
      '' multiply two matrices
    return( mat3( _
         A.a * B.a + A.b * B.d + A.c * B.g, A.a * B.b + A.b * B.e + A.c * B.h, A.a * B.c + A.b * B.f + A.c * B.i, _
         A.d * B.a + A.e * B.d + A.f * B.g, A.d * B.b + A.e * B.e + A.f * B.h, A.d * B.c + A.e * B.f + A.f * B.i, _
         A.g * B.a + A.h * B.d + A.i * B.g, A.g * B.b + A.h * B.e + A.i * B.h, A.g * B.c + A.h * B.f + A.i * B.i ) )
   end operator
   
   operator * ( byref A as mat3, byref v as vec2h ) as vec2h export
      '' multiply a matrix with a column vector, resulting in a column vector
      return( vec2h( A.a * v.x + A.b * v.y + A.c * v.w, A.d * v.x + A.e * v.y + A.f * v.w, A.g * v.x + A.h * v.y + A.i * v.w ) ) 
   end operator
   
   operator * ( byref v as vec2h, byref A as mat3 ) as vec2h export
      '' multiply a vector with a row matrix, resulting in a row vector
      return( vec2h( A.a * v.x + A.d * v.y + A.g * v.w, A.b * v.x + A.e * v.y + A.h * v.w, A.c * v.x + A.f * v.y + A.i * v.w ) )
   end operator

   operator / ( byref A as mat3, byref s as double ) as mat3 export
      '' divide a matrix trough a scalar
      return( mat3( _
         A.a / s, A.b / s, A.c / s, _
         A.d / s, A.e / s, A.f / s, _
         A.g / s, A.h / s, A.i / s ) )
   end operator
   
   namespace matrices
      function identity2h() as mat3
         '' returns a 2D homogeneous identity matrix
         return( mat3( _
            1.0, 0.0, 0.0, _
            0.0, 1.0, 0.0, _
            0.0, 0.0, 1.0 ) )
      end function
            
      function translation2h( byval tx as double, byval ty as double ) as mat3
         '' returns a 2D homogeneous translation matrix
         return( mat3( _
            1.0, 0.0, tx, _
            0.0, 1.0, ty, _
            0.0, 0.0, 1.0 ) )
      end function
      
      function scaling2h( byval sx as double, byval sy as double ) as mat3
         '' returns a 2D scaling matrix
         return( mat3( _
            sx,      0.0,   0.0, _
            0.0,   sy,      0.0, _
            0.0,   0.0,   1.0 ) )
      end function
      
      function shear2h( byval sx as double, byval sy as double ) as mat3
         '' returns a shearig matrix by sx and sy angles in degrees
         return( mat3( _
            1.0,                            tan( radians( sx ) ), 0.0, _
            tan( radians( sy ) ), 1.0,                            0.0, _
            0.0,                            0.0,                            1.0 ) )   
      end function
      
      function rotation2h( byval ang as double ) as mat3
         '' returns a 2D rotation matrix around by ang degrees about the origin
         dim as double co = cos( radians( ang ) )
         dim as double si = sin( radians( ang ) )
         
         return( mat3( _
            co   , -si   ,   0.0, _
            si   , co   ,   0.0, _
            0.0   ,   0.0   ,   1.0 ) )
      end function
   
      function rotationAxis2h( byval axis as vec2h, byval ang as double ) as mat3
         '' returns a 2D rotation matrix around an arbitrary axis, by ang degrees
         dim as double co = cos( radians( ang ) )
         dim as double si = sin( radians( ang ) )

         return( mat3( _
            co, -si, -axis.x * co + axis.y * si + axis.x, _
            si,  co, -axis.x * si - axis.y * co + axis.y, _
            0.0, 0.0, 1.0 ) )
      end function
   end namespace
#endif
